/*-------------------------------------------------------------------*/
/* PNNGBB3.C          Pasi Frnti & Olli Virmajoki                   */
/*                                                                   */
/* Fast PNN method / Global Optimization                             */ 
/*   Branch and Bound / Variant bb3                                  */
/*   Look-ahead optimization                                         */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName	   "PNN-GBB3"
#define VersionNumber  "Version 0.4"
#define LastUpdated    "1.10.2000" 
/*-------------------------------------------------------------------*/

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <time.h>
#include <values.h>

#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "sort.h"
#include "solution.h"
#include "pnnfast.h"


/*---------------------------  M i s c	-----------------------------*/


/*#define  STATISTICS 1*/
#ifdef STATISTICS
#define STATCODE(a)a
#else 
#define STATCODE(a)
#endif

#define  round(a)		((a)+0.5)
#define  roundpds(a)    ((a)+0.9999999)
#define  swap(a,b)      { int t = a; a = b; b = t; }

static llong cfsol=0LL;
static llong csol=0LL;

/*===================================================================*/


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s=t%60; t=t/60; m=t%60; h=t/60;
  printf("  %li:%02li:%02li\n", h, m, s);
  fflush(stdout);
}


/*-------------------------------------------------------------------*/


/* ========================  OPT  ================================== */


static void CopyNN(NEARESTTYPE NNsource[], NEARESTTYPE NNdest[], int booksize)
{
    int i;
    for (i=0; i<booksize; i++)
    {
        NNdest[i].VectorDist = NNsource[i].VectorDist;
        NNdest[i].Vector = NNsource[i].Vector;
        NNdest[i].Recalculate = NNsource[i].Recalculate;
    }
}


static void CreateDuplicateState(TRAININGSET*   TS,
                                 CODEBOOK*      CB,
                                 PARTITIONING*  P,
                                 NEARESTTYPE    NN[],
                                 CODEBOOK*      CB2,
                                 PARTITIONING*  P2,
                                 NEARESTTYPE**  NN2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    *NN2 = (NEARESTTYPE*) allocate(BookSize(CB) * sizeof(NEARESTTYPE));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
    CopyNN(NN, *NN2, BookSize(CB));
}


static void FreeDuplicateState(CODEBOOK*     CB2,
                               PARTITIONING* P2,
                               NEARESTTYPE   NN2[])                             
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
    deallocate(NN2);
}


static void CreateDuplicateSolution(TRAININGSET*    TS,
                                    CODEBOOK*       CB,
                                    PARTITIONING*   P,
                                    CODEBOOK*       CB2,
                                    PARTITIONING*   P2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
}


static void FreeDuplicateSolution(CODEBOOK*    CB2,
                                  PARTITIONING* P2)
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
}

static void Depthsearch(TRAININGSET*  TS,
                        CODEBOOK*     CB,
                        PARTITIONING* P,
                        NEARESTTYPE   NN[],
                        int           FinalSize,
                        CODEBOOK*     CBOpt,
                        PARTITIONING* POpt,
                        double*       ErrorOpt,
                        CODEBOOK*     CBI,
                        PARTITIONING* PI,
                        CODEBOOK*     CBOptI,
                        PARTITIONING* POptI,
                        int           NewSizeI,
                        int           Quiet,
                        int           Pds,
                        int           Mps,
                        int           Lazy)
{
    CODEBOOK CB2;
    PARTITIONING P2;
    NEARESTTYPE* NN2;
    int i, j, S;
    double Error=0;
      
      S = BookSize(CB)-1;
      for( i=0; i < S; i++)
        for( j = i+1; j <= S; j++)
        {
            /* We create duplicate state: CB -> CB2 & P -> P2 */
            CreateDuplicateState(TS, CB, P, NN, &CB2, &P2, &NN2);
            if(Quiet >= 2)
            {
            printf("\nBookSize = %d ",BookSize(&CB2));
            }
            MergeVectors(TS, &CB2, &P2, NN2, i, j);
            csol+=1;
            /*RecalculateDistances(&CB2, NN2, i, Quiet, Pds, Lazy);*/
            if(BookSize(&CB2)==NewSizeI)
            {
                CopyCodebook(&CB2, CBI);
                CopyPartitioning(&P2, PI);
            }
            if(BookSize(&CB2) > FinalSize)
            {
                /* Checking the Bound -condition: */
                Error = AverageErrorCBFast(TS, &CB2, &P2, MSE);
                if(Error < *ErrorOpt)
                Depthsearch(TS, &CB2, &P2, NN2, FinalSize, CBOpt, POpt, ErrorOpt, CBI, PI, CBOptI, POptI, NewSizeI, 
                        Quiet, Pds, Mps, Lazy);
            }
            else
            {
                Error = AverageErrorCBFast(TS, &CB2, &P2, MSE);
                cfsol+=1;
                if(Error < *ErrorOpt)
                {
                    *ErrorOpt = Error;
                    CopyCodebook(&CB2, CBOpt);
                    CopyPartitioning(&P2, POpt);
                   
                    CopyCodebook(CBI, CBOptI);
                    CopyPartitioning(PI, POptI);
                }
                if(Quiet >= 2)
                printf("\nFinalSize= %d ",BookSize(&CB2));
            }
            /* We restore original state: CB2 -> CB & P2 -> P */
            FreeDuplicateState(&CB2, &P2, NN2);
        }
}


/* ================================================================= */


/*-------------------------------------------------------------------*/


double PNNGBb3(TRAININGSET*	    TS,
			   CODEBOOK*		CB,
			   PARTITIONING*	P,
               int              FinalSize,
               int              StepSize,
               int              Quiet,
               long             watch,
               int              Pds,
               int              Mps,
               int              Lazy)
{
  CODEBOOK CBOpt,CBOptI, CBI;
  PARTITIONING POpt, POptI, PI;
  NEARESTTYPE*  NN;
  DATA D;
  int NewSize, NewSizeI;
  double ErrorIni = MAXLLONG;
  double *ErrorOpt = &ErrorIni;

  RemoveEmptyPartitions(TS, CB, P, FinalSize);
  CreateDataStructures(CB, &D, &NN, Mps);
  InitializeData(CB, &D, NN, Mps);
  /*CalculateAllDistances(CB, &D, NN, Quiet, watch, Pds, Mps);*/
  Pds = 0;
  Mps = 0;
  Lazy = 0;

  if(BookSize(CB) > FinalSize)
  {

  while(BookSize(CB) > FinalSize)
  {
  NewSize = BookSize(CB) - StepSize;
  if(NewSize < FinalSize)
      NewSize = FinalSize;
  NewSizeI = BookSize(CB)-1; 

  CreateDuplicateSolution(TS, CB, P, &CBOpt, &POpt);
  CreateDuplicateSolution(TS, CB, P, &CBI, &PI);
  CreateDuplicateSolution(TS, CB, P, &CBOptI, &POptI);

  PNNFast(TS, &CBOptI, &POptI, NewSizeI, Quiet, watch, Pds, Mps, Lazy);
  PNNFast(TS, &CBOpt, &POpt, NewSize, Quiet, watch, Pds, Mps, Lazy);
  (*ErrorOpt) = AverageErrorCBFast(TS, &CBOpt, &POpt, MSE);

  Depthsearch(TS, CB, P, NN, NewSize, &CBOpt, &POpt, ErrorOpt, &CBI, &PI, &CBOptI, &POptI, NewSizeI, Quiet, Pds, Mps, Lazy);

  CopyCodebook(&CBOptI, CB);
  CopyPartitioning(&POptI, P);

  FreeDuplicateSolution(&CBOpt, &POpt);
  FreeDuplicateSolution(&CBI, &PI);
  FreeDuplicateSolution(&CBOptI, &POptI);

  printf("Booksize: %d \n",BookSize(CB));
  }
  
  }
  else
  {
  (*ErrorOpt) = AverageErrorCBFast(TS, CB, P, MSE);
  }

  FreeDataStructures(&D, NN, Mps);
  printf("Final solutions: %7.0f All solutions: %7.0f \n",(double)cfsol,(double)csol);
  return (*ErrorOpt);
}

