/*-------------------------------------------------------------------*/
/* PNNGBB4.C          Pasi Frnti & Olli Virmajoki                   */
/*                    & Timo Kaukoranta                              */
/*                                                                   */
/* Fast PNN method / Global Optimization                             */
/*   Branch and Bound / Variant duplicate free ; simple termination  */
/*                                                                   */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "PNN-GBB4"
#define VersionNumber  "Version 0.2"
#define LastUpdated    "11.11.2000"

/*-------------------------------------------------------------------*/

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <time.h>
#include <values.h>

#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "pnnfast.h"


/*---------------------------  M i s c  -----------------------------*/


/*#define  STATISTICS 1*/
#ifdef STATISTICS
#define STATCODE(a)a
#else
#define STATCODE(a)
#endif

#define  round(a)       ((a)+0.5)
#define  roundpds(a)    ((a)+0.9999999)
#define  swap(a,b)      { int t = a; a = b; b = t; }

static llong cfsol=0LL;
static llong csol=0LL;

/*===================================================================*/


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s=t%60; t=t/60; m=t%60; h=t/60;
  printf("  %li:%02li:%02li\n", h, m, s);
  fflush(stdout);
}


/* ========================  OPT  ================================== */

/*
static void CopyNN(NEARESTTYPE NNsource[], NEARESTTYPE NNdest[], int booksize)
{
    int i;
    for (i=0; i<booksize; i++)
    {
        NNdest[i].VectorDist = NNsource[i].VectorDist;
        NNdest[i].Vector = NNsource[i].Vector;
        NNdest[i].Recalculate = NNsource[i].Recalculate;
    }
}
*/

static void CreateDuplicateState(TRAININGSET*   TS,
                                 CODEBOOK*      CB,
                                 PARTITIONING*  P,
                                 CODEBOOK*      CB2,
                                 PARTITIONING*  P2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
}


static void FreeDuplicateState(CODEBOOK*     CB2,
                               PARTITIONING* P2)
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
}


static void CreateDuplicateSolution(TRAININGSET*    TS,
                                    CODEBOOK*       CB,
                                    PARTITIONING*   P,
                                    CODEBOOK*       CB2,
                                    PARTITIONING*   P2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
}


static void FreeDuplicateSolution(CODEBOOK*    CB2,
                                  PARTITIONING* P2)
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
}


static void Depthsearch(TRAININGSET*  TS,
                        CODEBOOK*     CB,
                        PARTITIONING* P,
                        int           FinalSize,
                        CODEBOOK*     CBOpt,
                        PARTITIONING* POpt,
                        double*       ErrorOpt,
                        int           Quiet,
                        int           a,
                        int           b)
{
  int i, j;
  CODEBOOK     CB2;
  PARTITIONING P2;
  double Error;

  for( i=a; i<FinalSize; i++ )
    {
    for( j=i+1; j<BookSize(CB); j++ )
      {
      if(Quiet >= 2)
        {
        printf("\nBookSize=%d (%d,%d) ",BookSize(CB),i,j);
        }
      CreateDuplicateState(TS, CB, P, &CB2, &P2);

      /* Merge */
      JoinPartitions(TS, &P2, i, j);
      PartitionCentroid(&P2, i, &Node(&CB2, i));
      CopyNode(&Node(&CB2, BookSize(&CB2)-1), &Node(&CB2, j), VectorSize(&CB2));
      BookSize(&CB2)--;
      csol+=1;

      /* Checking the Bound -condition: */
      Error = AverageErrorCBFast(TS, &CB2, &P2, MSE);
      if(Quiet >= 3)
        {
        printf("%9.4f vs. %9.4f ", Error, *ErrorOpt);
        }
      if( Error < *ErrorOpt )
        {
        if( BookSize(&CB2) == FinalSize )
          {
          *ErrorOpt = Error;
          CopyCodebook(&CB2, CBOpt);
          CopyPartitioning(&P2, POpt);
          cfsol+=1;
          }
        else
          {
          Depthsearch(TS, &CB2, &P2, FinalSize, CBOpt, POpt, ErrorOpt, Quiet, i, j);
          }
        }
      FreeDuplicateState(&CB2, &P2);

      if(Quiet >= 2)
        {
        printf("DONE");
        }
      }
    }
}


/* ================================================================= */


double PNNGBb4(TRAININGSET*   TS,
               CODEBOOK*      CB,
               PARTITIONING*  P,
               int            FinalSize,
               int            Quiet,
               long           watch,
               int            Pds,
               int            Mps,
               int            Lazy)
{
  CODEBOOK      CBOpt;
  PARTITIONING  POpt;
  double        ErrorOpt;

  if( BookSize(CB) > FinalSize )
    {
    CreateDuplicateSolution(TS, CB, P, &CBOpt, &POpt);
    PNNFast(TS, &CBOpt, &POpt, FinalSize, Quiet, watch, Pds, Mps, Lazy);
    ErrorOpt = AverageErrorCBFast(TS, &CBOpt, &POpt, MSE);
    printf("\n%7.4f",ErrorOpt);
    PrintTime(watch);

/* ??? */    RemoveEmptyPartitions(TS, CB, P, FinalSize);

    Depthsearch(TS, CB, P, FinalSize, &CBOpt, &POpt, &ErrorOpt, Quiet, 0, 1);

    CopyCodebook(&CBOpt, CB);
    CopyPartitioning(&POpt, P);

    FreeDuplicateSolution(&CBOpt, &POpt);
    }
  else
    {
    ErrorOpt = AverageErrorCBFast(TS, CB, P, MSE);
    }
  printf("Final solutions: %7.0f All solutions: %7.0f \n",(double)cfsol,(double)csol);
  return( ErrorOpt );
}

/*
Pseudokoodi:
(Jos ette lyd yhtn bugia tuosta, niin olen yllttynyt.)

S: Solution i.e. clustering, codevectors, training vectors etc.
S.M: Current number of clusters in solution S
X: Number of merges
M: Desired number of clusters

PNNbrute(S, X, M)
{
  dmin = infinity;
  for( i=0; i<S.M-1 - (X-1); i++ )
    {
    for( j=i+1; j<S.M - (X-1); j++
      {
      MergeTwoAndCheckWhatHappens(S, X, M, i, j, dmin);
      }
    }
}

MergeTwoAndCheckWhatHappens(S, X, M, a, b, dmin)
{
  S.merge(a, b);
  X--;
  if( S.distortion >= dmin )
    {
    return;
    }
  else
    {
    if( X == 0 || S.M == M )
      {
      if( S.distortion < dmin)
        {
        dmin = S.distortion;
        }
      return;
      }
    else
      {
      for( i=a; i< S.M-1 - (X-1); i++ )
        {
        for( j=b; j<S.M - (X-1); j++
          {
          MergeTwoAndCheckWhatHappens(S, X, M, i, j, dmin);
          }
        }
      }
    }
}
*/

/*
Toinen muotoilu samalle menetelmlle.
Kutsu PNNbrute(S, X, M, 0, 1, dmin)

PNNbrute(S, X, M, a, b, dmin)
{

  for( i=a; i<S.M-1 - (X-1); i++ )
    {
    for( j=b; j<S.M - (X-1); j++
      {
      S' = S;
      S'.merge(i, j);
      X--;
      if( S'.distortion < dmin )
        {
        if( X == 0 || S'.M == M )
          {
          dmin = S'.distortion;
          }
        else
          {
          PNNbrute(S', X, M, i, j, dmin);
          }
        }
      }
    }
}
*/

